(eval-when-compile (require 'cl))

(defvar emacs-process-mode-map
  (let ((m (make-sparse-keymap)))
    (define-key m "\C-k" 'emacs-process-delete)
    (define-key m "g" 'emacs-process-list)
    m))

(defun emacs-process-delete (&optional name)
  (interactive (list (let ((def (get-text-property (point) 'process-name)))
                       (completing-read (if def
                                            (format "Delete process (default %s): " def)
                                            "Delete process: ")
                                        (mapcar 'process-name (process-list))
                                        nil t nil nil def))))
  (delete-process (get-process name)))

(define-derived-mode emacs-process-mode fundamental-mode "Emacs Processes"
  (setq buffer-read-only t))

(defun emacs-process-info (&optional query-only)
  "Return a list of (NAMES STATUSES BUFFERS COMMANDS)."
  (let (processes statuses buffers commands)
    (mapc (lambda (p)
            (when (or (not query-only)
                      (and query-only (process-query-on-exit-flag p)))
              (push (process-name p) processes)
              (push (symbol-name (process-status p)) statuses)
              (push (let ((buf (process-buffer p)))
                      (cond
                       ((null buf) "(none)")
                       ((not (buffer-live-p buf)) "(Killed)")
                       (t (buffer-name buf))))
                    buffers)
              (push (case (process-type p)
                      (network (destructuring-bind
                                   (&key server type service host
                                         &allow-other-keys)
                                   (process-contact p t)
                                 (format "(network %s %s)"
                                         (if type "datagram" "stream")
                                         (if server
                                             (format "server on %s" service)
                                           (format "connection to %s" host)))))
                      (serial (destructuring-bind
                                  (&key port speed &allow-other-keys)
                                  (process-contact p t)
                                (format "(serial port %s%s)"
                                        (or port "?")
                                        (if speed
                                            (format " at %s b/s" speed)
                                          ""))))
                      (otherwise (mapconcat 'identity (process-command p) " ")))
                    commands)))
          (process-list))
    (list processes statuses buffers commands)))

(defun emacs-process-list (&optional query-only)
  (interactive)
  (let ((buf (get-buffer-create "*Process List2*")))
    (with-current-buffer buf
      (let* ((inhibit-read-only t)
             (minwidth 9)
             (info (emacs-process-info query-only))
             (colwidth (mapcar (lambda (i)
                                 (max minwidth
                                      (1+ (apply 'max (mapcar 'length i)))))
                               info)))
        (erase-buffer)
        (destructuring-bind (names statuses buffers commands)
            info
          (loop for name in (append '("Proc" "----") names)
                for status in (append '("Status" "------") statuses)
                for buffer in (append '("Buffer" "------") buffers)
                for command in (append '("Command" "-------") commands)
                for index from 0
                ;; compute the indentation
                with (col1 col2 col3) = colwidth
                with ind1 = col1
                with ind2 = (+ ind1 col2)
                with ind3 = (+ ind2 col3)
                do
                (insert name)   (indent-to ind1)
                (insert status) (indent-to ind2)
                (insert buffer) (indent-to ind3)
                (insert command)
                (when (> index 1)
                  (put-text-property (line-beginning-position)
                                     (line-end-position)
                                     'process-name name))
                (insert "\n"))))
      (set-buffer-modified-p nil)
      (goto-char (point-min))
      (emacs-process-mode))
    (display-buffer buf)))

;; (defalias 'list-processes 'emacs-process-list)
